"""
The command line interface to adding entire days to the work times.
"""

# Copyright 2008, Markus Demleitner.  This file is under the GPL,
# see COPYING for details.

from datetime import date
import sys

from pftf import due
from pftf import settings
from pftf import timefile


class Error(settings.Error):
	pass


def completeDate(dateSpec):
	"""returns a date object from dateSpec.

	dateSpecs may be <day>, <month>-<day>, <year>-<month>-<day>.
	"""
	parts = dateSpec.split("-")
	labels = ["day", "month", "year"]
	try:
		parts = [int(p) for p in parts]
		if len(parts)>3:
			raise ValueError("Too many parts")
		parts.reverse()
		today = date.today()
		resParts = {"day": today.day, "month": today.month, "year": today.year}
		resParts.update(dict(list(zip(labels, parts))))
		return date(**resParts)
	except ValueError:
		raise Error("Invalid specification for date: '%s'"%dateSpec)


def parseCommandLine():
	"""parses the command line for pineday.
	"""
	from optparse import OptionParser
	parser = OptionParser(usage="%prog [options] <label> <date> [<end-date>]\n"
		"  adds day(s) to the work times.  Dates may leave out months and\n"
		"  years for their current values and must otherwise be in ISO format,\n"
		"  e.g., 15, 11-14, 2000-11-14")
	parser.add_option("-t", "--time", help="Register HOURS work time"
		" rather than usual daily work hours", type="int", metavar="HOURS",
		dest="hours", default=settings.get("hoursPerDay"))
	opts, args = parser.parse_args()
	if len(args) not in set([2,3]):
		parser.print_help()
		sys.exit(1)
	label = args[0]
	if not label in settings.get("wClasses"):
		raise Error("Invalid label '%s'.  Valid labels are %s"%(
			label, ", ".join(settings.get("wClasses"))))
	startDate = completeDate(args[1])
	if len(args)==3:
		endDate = completeDate(args[2])
	else:
		endDate = startDate
	return opts, label, startDate, endDate


def enterDays(opts, label, startDate, endDate):
	tf = timefile.TimeFile()
	for day in due.iterWorkdaysBetween(startDate, endDate):
		tf.append("%s\t%s\t%s"%(day.isoformat(), label, opts.hours))


def main():
	"""processes command line a call.
	"""
	try:
		enterDays(*parseCommandLine())
	except Error as msg:
		sys.stderr.write("%s: %s\nTry %s --help\n"%(sys.argv[0], msg, sys.argv[0]))
		sys.exit(1)
