"""
A class providing synchronized access to the time keeping file and
a parser for its content.
"""

# Copyright 2008, Markus Demleitner.  This file is under the GPL,
# see COPYING for details.

from datetime import date, datetime
import errno
import io
import os
import sys
import time

from pftf import settings


class TimeoutReached(settings.Error):
	"""will be raised if we cannot acquire a lock.
	"""


class LockMixin(object):
	"""is a mixin that provides lock and unlock methods.

	It requires the class mixing in to provide a lockName attribute
	with a path of the lock file.

	The class is based on symlink's atomicity.  We link to our proc
	entry, so by following the link, you might get some guess if the
	lock is stale (on linux...).

	We don't do that, though.
	"""
	def lock(self, timeout=10):
		"""locks the object.

		timeout is the maximum time we spend trying to acquire the lock.
		If a timeout expires, you will get a TimoutReached exception.
		"""
		linkTarget = "/proc/%d"%os.getpid()
		giveUpAt = time.time()+timeout
		while time.time()<giveUpAt:
			try:
				os.symlink(linkTarget, self.lockName)
			except (os.error, IOError) as msg:
				if msg.errno!=errno.EEXIST:
					raise    # there's something seriously wrong
				time.sleep(0.1)
			else:
				break
		else:
			raise TimeoutReached("File locked.  Remove %s if the lock is "
			" stale"%self.lockName)

	def unlock(self):
		if os.path.islink(self.lockName):
			if os.readlink(self.lockName):
				os.unlink(self.lockName)
			else:
				sys.stderr.write("pining: Lock is held by someone else."
					"  Not unlocking.\n"
					"If you think there's something wrong, remove %s.\n"%self.lockName)


class TimeFile(LockMixin):
	"""is an accessor to the central time file.

	You should only use this class to access the file.  Concurrent
	access is quite possible here, and so the locking enforced here
	may be important.

	Also, we try to minimize damage done should the file be edited
	using a text editor (which is "almost allowed").
	"""
	basename = settings.get("filenames", "piningFile")

	def __init__(self):
		self.fName = os.path.join(settings.dataDir, self.basename)
		self.lockName = os.path.join(settings.dataDir, "lock_"+self.basename)
		if not os.path.exists(self.fName):
			open(self.fName, "wb").close()

	def _decorateFile(self, f):
		"""amends f's close method with a call to our unlock and returns it.
		"""
		parent = self

		class UnlockingFile(object):
			def __init__(self, fileInstance):
				self.f = fileInstance

			def close(self):
				parent.unlock()
				return self.f.close()

			def __iter__(self):
				return iter(self.f)

			def __getattr__(self, name):
				return getattr(self.f, name)

		return UnlockingFile(f)

	def getAppendingFile(self, timeout=10):
		self.lock(timeout)
		return self._decorateFile(open(self.fName, "ab"))

	def getEmptiedFile(self, timeout=10):
		self.lock(timeout)
		return self._decorateFile(open(self.fName, "wb"))

	def getReadingFile(self, timeout=10):
		self.lock(timeout)
		return self._decorateFile(open(self.fName, "rb"))

	def getReadingStringFile(self, timeout=10):
		return io.TextIOWrapper(
			self.getReadingFile(timeout),
			encoding="utf-8")

	def append(self, line, timeout=10):
		f = self.getAppendingFile(timeout)
		try:
			f.write((line+"\n").encode("utf-8"))
		finally:
			f.close()

	def getLines(self):
		return self.read().decode("utf-8").split("\n")

	def read(self):
		f = self.getReadingFile()
		try:
			return f.read().decode("utf-8")
		finally:
			f.close()

	def getLastLine(self):
		"""returns the last line in the WTF or None if that's not possible for
		some reason.
		"""
		try:
			f = self.getReadingFile(timeout=0.1)
			f.seek(-min(os.path.getsize(self.fName), 80), 2)
			stuff = f.read().decode("utf-8")
			f.close()
			return stuff.split("\n")[-2]
		except: # we don't want this to raise any exceptions since it should
			pass  # never be critical
		return None

	def getTimestamp(self):
		return os.path.getmtime(self.fName)


class ArchiveFile(TimeFile):
	basename = settings.get("filenames", "archiveFile")


"""
Parsing the line lists is done using a simple state machine with four 
input tags:

ENDTIME -- a line of the format <isodatetime>
STARTTIME -- a line of the format <isodatetime><tab><label>
DAY -- a line for the format <isodate><tab><label>
EOF -- end of input
"""


# Tags
ENDTIME, STARTTIME, DAY, EOF = list(range(4))


class ParseError(settings.Error):
	pass


def makeStamp():
	return time.strftime(settings.get("isoformat"), time.localtime())


def lexDTLine(label, parts):
	"""returns a tag, payload tuple for start/stop times.
	
	It will raise a ValueError if we are not dealing with a line containing
	an ISO Datetime.
	"""
	try:
		timeSpec = datetime(*(time.strptime(parts[0], 
			settings.get("isoformat"))[0:6]))
	except ValueError:
		return None
	if len(parts)>2:
		raise ValueError("Too many tabs")
	if label is None:
		return ENDTIME, (timeSpec,)
	else:
		return STARTTIME, (timeSpec, label)


def lexDateLine(label, parts):
	"""lexes a line with a single date.
	"""
	timeSpec = date(*(time.strptime(parts[0], 
		settings.get("dateformat"))[0:3]))
	if label is None:
		raise ValueError("No label given for day")
	if len(parts)>3:
		raise ValueError("Too many tabs")
	if len(parts)>2:
		seconds = float(parts[2])*3600
	else:
		seconds = settings.get("hoursPerDay")*3600
	return DAY, (timeSpec, label, seconds)


def lexLine(inLine):
	"""returns a tag, payload tuple for inLine.

	Payload is a tuple, the first item of which always is a datetime;
	further items depend on the tag.

	This may raise a ValueError if the line format is bad.
	"""
	if not inLine:
		return EOF, (datetime.today(),)
	if settings.get("allowBlanksInLabels"):
		parts = inLine.split('\t')
	else:
		parts = inLine.split()
	timeSpec, label = None, None
	if len(parts)>1:
		label = parts[1]
	res = lexDTLine(label, parts)
	if res is None:  # lexDTLine declared itself not responsible
		res = lexDateLine(label, parts)
	return res


class LineListParser(object):
	"""is a finite-state parser for the times list format.

	Feed a list of lines using feedList, and fetch the result
	(suitable as an argument to times.Times' constructor) from the
	intervals attribute.
	"""
	def __init__(self):
		self.intervals, self._line = [], 1
		self._state = "waitInter"

	def _process_waitInter(self, tag, payload):
		if tag==ENDTIME:
			return "waitInter"
		elif tag==STARTTIME:
			self.curStart = payload
			return "inInter"
		elif tag==DAY:
			day, label, duration = payload
			self.intervals.append((datetime(*day.timetuple()[:6]), 
				duration, label))
			return "waitInter"
		elif tag==EOF:
			return "stopped"
		else:
			raise ParseError("Invalid tag %s"%tag)
	
	def _process_inInter(self, tag, payload):
		if tag==DAY:
			day, label, duration = payload
			self.intervals.append((datetime(*day.timetuple()[:3]), 
				duration, label))
			return "inInter"

		dt = payload[0]-self.curStart[0]
		if dt.days!=0:
			raise ParseError("Illegal work marathon at line %d, ending %s"%
				(self._line, payload))
		self.intervals.append((self.curStart[0], dt.seconds,
			self.curStart[1]))
		if tag==ENDTIME:
			del self.curStart
			return "waitInter"
		elif tag==STARTTIME:
			self.curStart = payload
			return "inInter"
		elif tag==EOF:
			del self.curStart
			return "stopped"
		else:
			raise ParseError("Invalid tag %d"%tag)

	def _process_stopped(self, tag, payload):
		raise ParseError("Feed after end of list in line %d"%self.line)

	def _process(self, lexed):
		return getattr(self, "_process_"+self._state)(*lexed)

	def finish(self):
		"""makes the parser go to the stopped state if it's not yet there.
		"""
		if self._state!="stopped":
			self.feed("")

	def feed(self, inLine):
		try:
			inSym = lexLine(inLine.strip())
		except ValueError as msg:
			raise ParseError("Invalid line '%s' at line %d (%s)"%(inLine.strip(), 
				self._line, str(msg)))
		self._state = self._process(inSym)
		self._line += 1
	
	def feedList(self, lineSource):
		for line in lineSource:
			self.feed(line)
		self.finish()


def parseLineList(lineSeq):
	p = LineListParser()
	p.feedList(lineSeq)
	return p.intervals
