"""
Structured representation of time spent.

The raw data is coming in in a file that's assumed to be sorted by time,
with an ISO timestamp, a tab, and a label.  If the label is empty, the
time until the next line is not counted against work time.
"""

# Copyright 2008, Markus Demleitner.  This file is under the GPL,
# see COPYING for details.

import itertools
import time
from datetime import date, datetime

from pftf import timefile
from pftf import settings


def normalized(dist):
	total = float(sum(dist.values()))
	return dict((k, v/total) for k,v in dist.items())


def iterateSource(src, predicate, designation):
	lineNo = 1
	try:
		for ln in src:
			if predicate(ln):
				yield ln
			lineNo += 1
	except timefile.ParseError as msg:
		raise timefile.ParseError("%s, line %d: %s"%(designation, lineNo, str(msg)))


def getSources(withArchive=False):
	sources = [(timefile.TimeFile().getReadingStringFile(), "Pining File")]
	if withArchive:
		sources.append((timefile.ArchiveFile().getReadingStringFile(),
			"Archive File"))
	return sources


class Times(object):
	"""is a structured representation of the times spent working on various
	things.

	It is constructed with a sequence of tuples (start, duration, label),
	where start is a datetime (local time) and duration is in seconds.

	Alternative constructors are given below.
	"""
	def __init__(self, times):
		self.times = list(times)
		self.dayIndex = None

	def getTransfer(self):
		"""returns the sum of all negative contributions in hours.

		(these are counted as transfers to the next month).
		"""
		transfers = 0
		for _, t, _ in self.times:
			if t<0:
				transfers += t
		return transfers/3600

	def getWorkTime(self):
		"""returns the total time worked for the data contained in self in hours.
		"""
		return sum(t[1] for t in self.times if t[1]>0)/3600.

	def getWorkTimes(self):
		dist = {}
		for _, t, label in self.times:
			if t>0:
				# negative times are transfers to the next month; they are not
				#	counted here but instead added to the due.
				dist[label] = dist.get(label, 0)+t
		return dist

	def getWorkTimeDist(self):
		return normalized(self.getWorkTimes())

	def _getDayIndex(self):
		if self.dayIndex is None:
			self.dayIndex = {}
			for t in self.times:
				if t[1]<0:
					# negative contributions count against the due (transfers)
					continue
				d = date(*t[0].timetuple()[:3])
				if d in self.dayIndex:
					self.dayIndex[d].append(t)
				else:
					self.dayIndex[d] = [t]
		return self.dayIndex

	def iterdays(self):
		dates = list(self._getDayIndex().keys())
		dates.sort()
		return iter(dates)

	def getForDay(self, date):
		return Times(self._getDayIndex()[date])

	@classmethod
	def fromLines(cls, lines):
		"""returns Times for lines.  Sequence matters.

		The function will raise a timefile.ParseError if any line is bad.
		"""
		p = timefile.LineListParser()
		for line in lines:
			p.feed(line)
		p.finish()
		return cls(p.intervals)

	@classmethod
	def forPredicate(cls, predicate, withArchive):
		sources = getSources(withArchive)
		source = itertools.chain(*[iterateSource(s, predicate, d) 
			for s, d in sources])
		try:
			res = cls.fromLines(source)
		finally:
			for f, _ in sources:
				f.close()
		return res

	@classmethod
	def forDateStart(cls, dateStart, withArchive=False):
		"""returns Times for all lines starting with dateStart.
		"""
		return cls.forPredicate(lambda ln: ln.startswith(dateStart), withArchive)

	@classmethod
	def forDates(cls, datesSet, withArchive=False):
		"""returns Times for all lines having dates in datesSet (which is
		a set of strings).

		This will only work if you keep the date format in ISO.
		"""
		return cls.forPredicate(lambda ln: ln[:10] in datesSet, withArchive)

	@classmethod
	def forMonth(cls, opts):
		"""returns Times for the month specified in opts (cf. commandline)
		"""
		return cls.forDateStart(
			date(opts.year, opts.month, opts.day).strftime(
				settings.get("monthformat")), withArchive=opts.withArchive)

	@classmethod
	def forToday(cls):
		"""returns Times for the current date out of the default log.
		"""
		return cls.forDateStart(date.today().isoformat())
